import yaml
from nanohttp import settings

from restfulpy.cli import RequireSubCommand, Launcher


class EmailLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('email', help='Sends a test email')
        parser.add_argument('target')
        parser.add_argument('-m', '--message', default='The email body')
        parser.add_argument('-s', '--subject', default='Test subject')
        parser.add_argument(
            '-f',
            '--from',
            dest='from_',
            default='restfulpy@example.com'
        )
        return parser

    def launch(self):
        from restfulpy.messaging import SmtpProvider

        p = SmtpProvider()
        p.send(
            self.args.target,
            self.args.subject,
            self.args.message,
            from_=self.args.from_
        )


class LogLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'log',
            help='Setup the server\'s database.')
        parser.add_argument(
            'message',
            nargs='+',
            help='The log message'
        )
        parser.add_argument(
            '-g',
            '--logger',
            default='main',
            help='The logger name, default: main'
        )
        parser.add_argument(
            '-l',
            '--level',
            default='debug',
            help='Log level, default info'
        )
        return parser

    def launch(self):
        from restfulpy.logging_ import get_logger
        getattr(
            get_logger(self.args.logger),
            self.args.level
        )(' '.join(self.args.message))


class ConfigurationDumpLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'config-dump',
            help='Dump the configuration'
        )
        parser.add_argument('path', help='The config path, for example: db')
        return parser

    def launch(self):
        dump = yaml.dump(
            getattr(settings, self.args.path),
            default_flow_style=False
        )
        print(dump)


class DevLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('dev', help="Development commands")
        dev_subparsers = parser.add_subparsers(
            title='admin command',
            dest='admin_command'
        )
        LogLauncher.register(dev_subparsers)
        EmailLauncher.register(dev_subparsers)
        ConfigurationDumpLauncher.register(dev_subparsers)
        return parser



from sqlalchemy import Integer, Unicode

from restfulpy.orm import DeclarativeBase, Field, ApproveRequiredMixin


class ApproveRequiredObject(ApproveRequiredMixin, DeclarativeBase):
    __tablename__ = 'approve_required_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


def test_approve_required_mixin(db):
    session = db(expire_on_commit=True)

    object1 = ApproveRequiredObject(
        title='object 1',
    )

    session.add(object1)
    session.commit()
    assert not object1.is_approved
    assert session.query(ApproveRequiredObject)\
        .filter(ApproveRequiredObject.is_approved).count() == 0

    object1.is_approved = True
    assert object1.is_approved
    session.commit()

    object1 = session.query(ApproveRequiredObject).one()
    assert  object1.is_approved

    json = object1.to_dict()
    assert 'isApproved' in json

    assert  session.query(ApproveRequiredObject)\
        .filter(ApproveRequiredObject.is_approved).count() == 1

    assert ApproveRequiredObject.filter_approved(session=session).count() == 1

    assert not ApproveRequiredObject.import_value(
        ApproveRequiredObject.is_approved, 'false'
    )
    assert not ApproveRequiredObject.import_value(
        ApproveRequiredObject.is_approved, 'FALSE'
    )
    assert not ApproveRequiredObject.import_value(
        ApproveRequiredObject.is_approved, 'False'
    )
    assert ApproveRequiredObject.import_value(
        ApproveRequiredObject.is_approved, 'true'
    )
    assert ApproveRequiredObject.import_value(
        ApproveRequiredObject.is_approved, 'TRUE'
    )
    assert ApproveRequiredObject.import_value(
        ApproveRequiredObject.is_approved, 'True'
    )

    assert ApproveRequiredObject.import_value(
        ApproveRequiredObject.title, 'title'
    ) == 'title'




from bddrest import response, status
from nanohttp import json
from sqlalchemy import Unicode, Integer

from restfulpy.controllers import ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import ApplicableTestCase


class Foo(DeclarativeBase):
    __tablename__ = 'foo'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(10))


class Root(ModelRestController):
    __model__ = Foo

    @json
    @commit
    @Foo.expose
    def post(self):
        m = Foo()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @Foo.expose
    def get(self):
        return DBSession.query(Foo).first()


class TestStringEncoding(ApplicableTestCase):
    __controller_factory__ = Root

    def test_string_codec(self):
        with self.given(
            'Title Has a backslash',
            verb='POST',
            form=dict(title='\\'),
        ):
            assert status == 200
            assert response.json['title'] == '\\'

        with self.given(
            'Getting foo',
            verb='GET',
        ):
            title = response.json['title']
            assert status == 200
            assert title == '\\'
            assert len(title) == 1



import argparse

from restfulpy.cli import RequireSubCommand, Launcher
from restfulpy.db import DatabaseManager
from restfulpy.orm import setup_schema


class BasedataLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser(
            'basedata',
            help='Setup the server\'s database.'
        )

    def launch(self):
        self.args.application.insert_basedata()


class MockupDataLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('mockup', help='Insert mockup data.')
        parser.add_argument('mockup_args', nargs=argparse.REMAINDER)
        return parser

    def launch(self):
        self.args.application.insert_mockup(self.args.mockup_args)


class CreateDatabaseLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'create',
            help='Create the server\'s database.'
        )
        parser.add_argument(
            '-d',
            '--drop',
            dest='drop_db',
            action='store_true',
            default=False,
            help='Drop existing database before create another one.'
        )
        parser.add_argument(
            '-s',
            '--schema',
            dest='schema',
            action='store_true',
            default=False,
            help='Creates database schema after creating the database.'
        )
        parser.add_argument(
            '-b',
            '--basedata',
            action='store_true',
            default=False,
            help= \
                'Implies `(-s|--schema)`, Inserts basedata after schema '
                'generation.'
        )
        parser.add_argument(
            '-m',
            '--mockup',
            action='store_true',
            default=False,
            help='Implies `(-s|--schema)`, Inserts mockup data.'
        )
        return parser

    def launch(self):
        with DatabaseManager() as db_admin:
            if self.args.drop_db:
                db_admin.drop_database()
            db_admin.create_database()
            if self.args.schema or self.args.basedata or self.args.mockup:
                setup_schema()
                if self.args.basedata:
                    self.args.application.insert_basedata()
                if self.args.mockup:
                    self.args.application.insert_mockup()


class DropDatabaseLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser(
            'drop',
            help='Drop the server\'s database.'
        )

    def launch(self):
        with DatabaseManager() as db_admin:
            db_admin.drop_database()


class CreateDatabaseSchemaLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser(
            'schema',
            help='Creates the database objects.'
        )

    def launch(self):
        setup_schema()


class DatabaseLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('db', help='Database administrationn')
        admin_subparsers = parser.add_subparsers(
            title='admin command',
            dest='admin_command'
        )
        CreateDatabaseSchemaLauncher.register(admin_subparsers)
        BasedataLauncher.register(admin_subparsers)
        MockupDataLauncher.register(admin_subparsers)
        DropDatabaseLauncher.register(admin_subparsers)
        CreateDatabaseLauncher.register(admin_subparsers)
        return parser




import ujson

from restfulpy.cli import Launcher, RequireSubCommand
from restfulpy.principal import JwtPrincipal


class JWTLauncher(Launcher, RequireSubCommand):  # pragma: no cover
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('jwt', help='JWT management')
        jwt_subparsers = parser.add_subparsers(
            title='JWT management',
            dest='jwt_command'
        )
        CreateJWTTokenLauncher.register(jwt_subparsers)
        return parser


class CreateJWTTokenLauncher(Launcher):  # pragma: no cover
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'create',
            help='Create a new initial jwt.'
        )
        parser.add_argument(
            '-e', '--expire-in',
            default=3600,
            type=int,
            help='the max age, default: 3600 (one hour).'
        )
        parser.add_argument(
            'payload',
            default='{}',
            nargs='?',
            help= \
                'A JSON parsable string to treat as payload. for example: '
                '{"a": "b"}'
        )
        return parser

    def launch(self):
        payload = ujson.loads(self.args.payload)
        print(JwtPrincipal(payload).dump(self.args.expire_in).decode())




import io
from os.path import dirname, abspath, join

from nanohttp import settings, configure

from restfulpy.messaging.providers import SmtpProvider
from restfulpy.mockup import mockup_smtp_server


HERE = abspath(dirname(__file__))


def test_smtp_provider():
    configure(force=True)
    settings.merge(f'''
        smtp:
          host: smtp.example.com
          port: 587
          username: user@example.com
          password: password
          local_hostname: localhost
          tls: false
          auth: false
          ssl: false
        messaging:
          mako_modules_directory: {join(HERE, '../../data', 'mako_modules')}
          template_dirs:
            - {join(HERE, 'templates')}
        ''',
    )

    with mockup_smtp_server() as (server, bind):
        settings.smtp.host = bind[0]
        settings.smtp.port = bind[1]

        # Without templates
        SmtpProvider().send(
            'test@example.com',
            'test@example.com',
            'Simple test body',
            cc='test@example.com',
            bcc='test@example.com'
        )

        # With template
        SmtpProvider().send(
            'test@example.com',
            'test@example.com',
            {},
            template_filename='test-email-template.mako'
        )

        # With attachments
        attachment = io.BytesIO(b'This is test attachment file')
        attachment.name = 'path/to/file.txt'
        SmtpProvider().send(
            'test@example.com',
            'test@example.com',
            'email body with Attachment',
            attachments=[attachment]
        )




from nanohttp import settings

from restfulpy.messaging import Email, create_messenger


def test_messaging_model(db):
    __configuration__ = '''
    messaging:
      default_sender: test@example.com
      default_messenger: restfulpy.mockup.MockupMessenger
    '''

    settings.merge(__configuration__)
    session = db()

    mockup_messenger = create_messenger()

    message = Email(
        to='test@example.com',
        subject='Test Subject',
        body={'msg': 'Hello'}
    )

    session.add(message)
    session.commit()

    message.do_({})

    assert mockup_messenger.last_message == {
        'body': {'msg': 'Hello'},
        'subject': 'Test Subject',
        'to': 'test@example.com'
    }




import re
from datetime import datetime

from nanohttp import context, HTTPBadRequest, settings
from sqlalchemy import DateTime, between, desc
from sqlalchemy.events import event
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql.expression import nullslast, nullsfirst

from .field import Field


FILTERING_IN_OPERATOR_REGEX = re.compile('!?IN\((?P<items>.*)\)')
FILTERING_BETWEEN_OPERATOR_REGEX = \
    re.compile('!?BETWEEN\((?P<min>.*),(?P<max>.*)\)')


class TimestampMixin:
    created_at = Field(
        DateTime,
        nullable=False,
        json='createdAt',
        readonly=True,
        default=datetime.utcnow,
    )


class ModifiedMixin(TimestampMixin):
    modified_at = Field(
        DateTime,
        nullable=True,
        json='modifiedAt',
        readonly=True
    )

    @property
    def last_modification_time(self):
        return self.modified_at or self.created_at

    @staticmethod
    def before_update(mapper, connection, target):
        target.modified_at = datetime.utcnow()

    @classmethod
    def __declare_last__(cls):
        event.listen(cls, 'before_update', cls.before_update)


class SoftDeleteMixin:
    removed_at = Field(
        DateTime,
        nullable=True,
        json='removedAt',
        readonly=True
    )

    def assert_is_not_deleted(self):
        if self.is_deleted:
            raise ValueError('Object is already deleted.')

    def assert_is_deleted(self):
        if not self.is_deleted:
            raise ValueError('Object is not deleted.')

    @property
    def is_deleted(self):
        return self.removed_at is not None

    def soft_delete(self, ignore_errors=False):
        if not ignore_errors:
            self.assert_is_not_deleted()
        self.removed_at = datetime.utcnow()

    def soft_undelete(self, ignore_errors=False):
        if not ignore_errors:
            self.assert_is_deleted()
        self.removed_at = None

    @staticmethod
    def before_delete(mapper, connection, target):
        raise AssertionError(f'Cannot remove {target}')

    @classmethod
    def __declare_last__(cls):
        event.listen(cls, 'before_delete', cls.before_delete)

    @classmethod
    def filter_deleted(cls, query):
        return query.filter(cls.removed_at.isnot(None))

    @classmethod
    def exclude_deleted(cls, query):
        return query.filter(cls.removed_at.is_(None))


class ActivationMixin:

    activated_at = Field(
        DateTime,
        nullable=True,
        json='activatedAt',
        readonly=True,
        protected=True
    )

    @hybrid_property
    def is_active(self):
        return self.activated_at is not None

    @is_active.setter
    def is_active(self, value):
        self.activated_at = datetime.utcnow() if value else None

    @is_active.expression
    def is_active(self):
        return self.activated_at.isnot(None)

    @classmethod
    def filter_activated(cls, query):
        return query.filter(cls.is_active)

    @classmethod
    def import_value(cls, column, v):
        if column.key == cls.is_active.key and not isinstance(v, bool):
            return str(v).lower() == 'true'
        return super().import_value(column, v)


class AutoActivationMixin(ActivationMixin):

    activated_at = Field(
        DateTime,
        nullable=True,
        json='activatedAt',
        readonly=True,
        protected=True,
        default=datetime.utcnow
    )


class DeactivationMixin(ActivationMixin):

    deactivated_at = Field(
        DateTime,
        nullable=True,
        json='deactivatedAt',
        readonly=True,
        protected=True
    )

    @ActivationMixin.is_active.setter
    def is_active(self, value):
        now = datetime.utcnow()
        if value:
            self.activated_at = now
            self.deactivated_at = None
        else:
            self.activated_at = None
            self.deactivated_at = now


class PaginationMixin:
    __take_header_key__ = 'HTTP_X_TAKE'
    __skip_header_key__ = 'HTTP_X_SKIP'
    __max_take__ = 100

    @classmethod
    def paginate_by_request(cls, query):

        try:
            take = int(
                context.query.get('take') \
                or context.environ.get(cls.__take_header_key__) \
                or cls.__max_take__
            )

            skip = int(
                context.query.get('skip') \
                or context.environ.get(cls.__skip_header_key__) \
                or 0
            )
        except ValueError:
            raise HTTPBadRequest()

        if take > cls.__max_take__:
            raise HTTPBadRequest()

        context.response_headers.add_header('X-Pagination-Take', str(take))
        context.response_headers.add_header('X-Pagination-Skip', str(skip))
        context.response_headers.add_header(
            'X-Pagination-Count',
            str(query.count())
        )
        return query.offset(skip).limit(take)


class FilteringMixin:
    @classmethod
    def filter_by_request(cls, query):

        for c in cls.iter_json_columns():
            json_name = cls.get_column_info(c)['json']
            if json_name in context.query:
                value = context.query[json_name]
                query = cls._filter_by_column_value(query, c, value)

        return query

    @classmethod
    def _filter_by_column_value(cls, query, column, value):

        def return_(e):
            return query.filter(e)

        import_value = getattr(cls, 'import_value')
        if not isinstance(value, str):
            raise HTTPBadRequest()

        in_operator_match = FILTERING_IN_OPERATOR_REGEX.match(value)
        if in_operator_match:
            not_ = value.startswith('!')
            items = in_operator_match.groupdict()['items'].split(',')
            items = [i for i in items if i.strip() != '']
            if not len(items):
                raise HTTPBadRequest('Invalid query string: %s' % value)
            expression = column.in_([import_value(column, j) for j in items])
            if not_:
                expression = ~expression

            return return_(expression)

        between_operator_match = FILTERING_BETWEEN_OPERATOR_REGEX.match(value)
        if between_operator_match:
            not_ = value.startswith('!')
            groups = between_operator_match.groupdict()
            start, end = groups['min'].strip(), groups['max'].strip()
            if not (start or end):
                raise HTTPBadRequest('Invalid query string: %s' % value)
            expression = between(column, start, end)
            if not_:
                expression = ~expression

            return return_(expression)

        if value == 'null':
            expression = column.is_(None)
        elif value == '!null':
            expression = column.isnot(None)
        elif value.startswith('!'):
            expression = column != import_value(column, value[1:])
        elif value.startswith('>='):
            expression = column >= import_value(column, value[2:])
        elif value.startswith('>'):
            expression = column > import_value(column, value[1:])
        elif value.startswith('<='):
            expression = column <= import_value(column, value[2:])
        elif value.startswith('<'):
            expression = column < import_value(column, value[1:])

        # LIKE
        elif '%' in value:
            func, actual_value = \
                (column.ilike, value[1:]) if value.startswith('~') \
                else (column.like, value)
            expression = func(import_value(column, actual_value))

        # EQUAL
        else:
            expression = column == import_value(column, value)

        return return_(expression)


class OrderingMixin:
    @classmethod
    def _sort_by_key_value(cls, query, column, descending=False):
        expression = column

        if descending:
            expression = desc(expression)

        if settings.db.url.startswith('sqlite'):
            return query.order_by(expression)

        return query.order_by(
            (nullsfirst if descending else nullslast)(expression)
        )

    @classmethod
    def sort_by_request(cls, query):
        sort_exp = context.query.get('sort', '').strip()
        if not sort_exp:
            return query

        sort_columns = [
            (
                c[1:] if c.startswith('-') else c,
                'desc' if c.startswith('-') else 'asc'
            )
            for c in sort_exp.split(',')
        ]

        criteria = cls.create_sort_criteria(sort_columns)

        for criterion in criteria:
            query = cls._sort_by_key_value(query, *criterion)

        return query


class ApproveRequiredMixin:
    approved_at = Field(
        DateTime,
        nullable=True,
        json='approvedAt',
        readonly=True
    )

    @hybrid_property
    def is_approved(self):
        return self.approved_at is not None

    @is_approved.setter
    def is_approved(self, value):
        self.approved_at = datetime.utcnow() if value else None

    @is_approved.expression
    def is_approved(self):
        return self.approved_at.isnot(None)

    @classmethod
    def import_value(cls, column, v):
        if column.key == cls.is_approved.key and not isinstance(v, bool):
            return str(v).lower() == 'true'
        return super().import_value(column, v)

    @classmethod
    def filter_approved(cls, query=None, session=None):
        if query is None:
            query = session.query(cls) if session else cls.query
        return query.filter(cls.is_approved)


class FullTextSearchMixin:
    __ts_vector__ = None

    @classmethod
    def search(cls, expressions, query):
        expressions = expressions.replace(' ', '|')
        return query.filter(
            cls.__ts_vector__.match(expressions)
        )




from datetime import date

import pytest
from bddrest import response, when, status
from dateutil.tz import tzoffset
from nanohttp import json
from sqlalchemy import Integer, Date

from restfulpy.configuration import settings
from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.datetimehelpers import parse_datetime, format_datetime, \
    parse_date, format_date
from restfulpy.mockup import mockup_localtimezone
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import ApplicableTestCase


class Party(DeclarativeBase):
    __tablename__ = 'party'
    id = Field(Integer, primary_key=True)
    when = Field(Date)


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = 'party'

    @json
    @commit
    def post(self):
        m = Party()
        m.update_from_request()
        DBSession.add(m)
        return m.when.isoformat()


class TestDate(ApplicableTestCase):
    __controller_factory__ = Root

    def test_update_from_request(self):
        with self.given(
            'Posting a date in form',
            verb='POST',
            form=dict(
                when='2001-01-01',
            )
        ):
            assert status == 200
            assert response.json == '2001-01-01'

            when(
                'Posix time format',
                form=dict(
                    when='1513434403'
                )
            )
            assert status == 200
            assert response.json == '2017-12-16'

            when(
                'Posting a datetime instead of date',
                form=dict(
                    when='2001-01-01T00:01:00.123456'
                )
            )
            assert status == 200
            assert response.json == '2001-01-01'

            when(
                'Posting an invalid datetime',
                form=dict(
                    when='2001-00-01'
                )
            )
            assert status == '400 Invalid date or time: 2001-00-01'

    def test_date_parsing(self):
        assert date(1970, 1, 1) == parse_date('1970-01-01')
        assert date(1970, 1, 1) == parse_date('1970-01-01T00:00:00.001')
        assert date(1970, 1, 1) == parse_date('1970-01-01T00:00:00.001000')

    def test_date_formatting(self):
        assert '1970-01-01' == format_date(date(1970, 1, 1))

    def test_unix_timestamp(self):
        assert date(1970, 1, 1) == parse_date('1.3343')




import argparse
import os
from os.path import join

from alembic.config import main as alembic_main

from restfulpy.cli import Launcher


class MigrateLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        migrate_parser = subparsers.add_parser(
            'migrate',
            help="Executes the alembic command"
        )
        migrate_parser.add_argument(
            'alembic_args',
            nargs=argparse.REMAINDER,
            help="For more information, please see `alembic --help`"
        )
        return migrate_parser

    def launch(self):
        current_directory = os.curdir
        try:
            os.chdir(self.args.application.root_path)
            alembic_ini = join(self.args.application.root_path, 'alembic.ini')
            alembic_main(
                argv=['--config', alembic_ini] + self.args.alembic_args
            )
        finally:
            os.chdir(current_directory)




import importlib.util
import re
import sys
import warnings
from hashlib import md5
from os.path import dirname, abspath


def import_python_module_by_filename(name, module_filename):
    """
    Import's a file as a python module, with specified name.

    Don't ask about the `name` argument, it's required.

    :param name: The name of the module to override upon imported filename.
    :param module_filename: The filename to import as a python module.
    :return: The newly imported python module.
    """

    sys.path.append(abspath(dirname(module_filename)))
    spec = importlib.util.spec_from_file_location(
        name,
        location=module_filename)
    imported_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(imported_module)
    return imported_module


def construct_class_by_name(name, *args, **kwargs):
    """
    Construct a class by module path name using *args and **kwargs

    Don't ask about the `name` argument, it's required.

    :param name: class name
    :return: The newly imported python module.
    """
    parts = name.split('.')
    module_name, class_name = '.'.join(parts[:-1]), parts[-1]
    module = importlib.import_module(module_name)
    return getattr(module, class_name)(*args, **kwargs)


def deprecated(func):  # pragma: no cover
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used.
    """

    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # Turn off filter
        warnings.warn(
            'Call to deprecated function %s.' % func.__name__,
            category=DeprecationWarning,
            stacklevel=2
        )
        warnings.simplefilter('default', DeprecationWarning)  # Reset filter
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


def to_camel_case(text):
    return re.sub("(_\w)", lambda x: x.group(1)[1:].upper(), text)


def to_pascal_case(text):
    return to_camel_case(text).capitalize()


def copy_stream(source, target, *, chunk_size: int= 16 * 1024) -> int:
    length = 0
    while 1:
        buf = source.read(chunk_size)
        if not buf:
            break
        length += len(buf)
        target.write(buf)
    return length


def md5sum(f):
    if isinstance(f, str):
        file_obj = open(f, 'rb')
    else:
        file_obj = f

    try:
        checksum = md5()
        while True:
            d = file_obj.read(1024)
            if not d:
                break
            checksum.update(d)
        return checksum.digest()
    finally:
        if file_obj is not f:
            file_obj.close()




from bddrest import response, status, when
from nanohttp import json
from sqlalchemy import Unicode, Integer

from restfulpy.controllers import ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import ApplicableTestCase


class Supervisor(DeclarativeBase):
    __tablename__ = 'supervisor'

    id = Field(
        Integer,
        primary_key=True,
        readonly='709 id is readonly'
    )
    email = Field(
        Unicode(100),
        not_none='701 email cannot be null',
        required='702 email required',
        pattern=(
            '(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)',
            '707 Invalid email address'
        ),
        watermark='Email',
        example="user@example.com"
    )
    title = Field(
        Unicode(50),
        index=True,
        min_length=(4, '703 title must be at least 4 characters'),
        max_length=(50, '704 Maximum allowed length for title is 50'),
        watermark='First Name',
        nullable=True,
    )
    age = Field(
        Integer,
        nullable=True,
        python_type=(int, '705 age must be integer'),
        minimum=(18, '706 age must be greater than 17'),
        maximum=(99, '706 age must be less than 100')
    )


class Root(ModelRestController):
    __model__ = Supervisor

    @json
    @commit
    @Supervisor.validate(strict=True)
    @Supervisor.expose
    def post(self):
        m = Supervisor()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @commit
    @Supervisor.validate
    @Supervisor.expose
    def put(self):
        m = Supervisor()
        m.update_from_request()
        DBSession.add(m)
        return m


    @json
    @Supervisor.expose
    def get(self):
        return DBSession.query(Supervisor).first()


class TestModelValidationDecorator(ApplicableTestCase):
    __controller_factory__ = Root

    def test_string_codec(self):
        with self.given(
            'Whitebox',
            verb='POST',
            form=dict(
                title='white',
                email='user@example.com',
            ),
        ):
            assert status == 200
            assert response.json['title'] == 'white'

            when(
                'Not strict',
                verb='PUT',
                form=dict(
                    email='userput@example.com'
                )
            )
            assert status == 200


            when(
                'Invalid email format',
                form=dict(
                    title='black1',
                    email='inavlidformat'
                )
            )
            assert status == 707

            when(
                'Passing readonly field',
                form=dict(
                    id=22,
                    title='black2',
                    email='user@example.com'
                )
            )
            assert status == 709




import pytest
from nanohttp import HTTPBadRequest
from nanohttp.contexts import Context
from sqlalchemy import Integer, Unicode

from restfulpy.orm import DeclarativeBase, Field, PaginationMixin


class PagingObject(PaginationMixin, DeclarativeBase):
    __tablename__ = 'paging_object'
    __max_take__ = 4

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


def test_pagination_mixin(db):
    session = db()

    for i in range(1, 6):
        obj = PagingObject(
            title='object %s' % i,
        )
        session.add(obj)
    session.commit()

    query = session.query(PagingObject)

    with Context({'QUERY_STRING': 'take=2&skip=1'}) as context:
        assert PagingObject.paginate_by_request(query).count() == 2
        assert context.response_headers['X-Pagination-Take'] == '2'
        assert context.response_headers['X-Pagination-Skip'] == '1'
        assert context.response_headers['X-Pagination-Count'] == '5'

    with Context({'QUERY_STRING': 'take=two&skip=one'}), \
        pytest.raises(HTTPBadRequest):
        PagingObject.paginate_by_request(query).count()

    with Context({'QUERY_STRING': 'take=5'}), pytest.raises(HTTPBadRequest):
        PagingObject.paginate_by_request(query)




import pytest
import ujson
from nanohttp import context, json, text, RestController
from nanohttp.contexts import Context

from restfulpy.controllers import JsonPatchControllerMixin


class BiscuitsController(RestController):
    @json
    def put(self, id_: int = None):
        result = {}
        result.update(context.form)
        result['id'] = id_
        result['a'] = context.query.get('a')
        return result

    @json
    def get(self, id_: int = None):
        result = {}
        result.update(context.form)
        result['id'] = id_
        return result

    @json
    def error(self):
        raise Exception()


class SimpleJsonPatchController(JsonPatchControllerMixin, RestController):
    biscuits = BiscuitsController()

    @text
    def get(self):
        yield 'hey'


def test_jsonpatch_rfc6902():
    environ = {
        'REQUEST_METHOD': 'PATCH'
    }
    with Context(environ):
        controller = SimpleJsonPatchController()
        context.form = [
            {
                'op': 'get',
                'path': '/'
            },
            {
                'op': 'put',
                'path': 'biscuits/1',
                'value': {'name': 'Ginger Nut'}
            },
            {
                'op': 'get',
                'path': 'biscuits/2',
                'value': {'name': 'Ginger Nut'}
            }
        ]
        result = ujson.loads(controller())
        assert len(result) == 3


def test_jsonpatch_error():
    environ = {
        'REQUEST_METHOD': 'PATCH'
    }
    with Context(environ), pytest.raises(Exception):
        controller = SimpleJsonPatchController()
        context.form = [
            {
                'op': 'put',
                'path': 'biscuits/1',
                'value': {'name': 'Ginger Nut'}
            },
            {
                'op': 'error',
                'path': 'biscuits',
                'value': None
            }
        ]

        controller()


def test_jsonpatch_querystring():
    environ = {
        'REQUEST_METHOD': 'PATCH',
        'QUERY_STRING': 'a=10'
    }
    with Context(environ):
        controller = SimpleJsonPatchController()
        context.form = [
            {
                'op': 'get',
                'path': '/'
            },
            {
                'op': 'put',
                'path': 'biscuits/1?a=1',
                'value': {'name': 'Ginger Nut'}
            },
            {
                'op': 'get',
                'path': 'biscuits/2',
                'value': {'name': 'Ginger Nut'}
            }
        ]
        result = ujson.loads(controller())
        assert len(result) == 3
        assert result[1]['a'] == '1'
        assert 'a' not in result[0]
        assert 'a' not in result[2]


def test_jsonpatch_caseinsesitive_verb():
    environ = {
        'REQUEST_METHOD': 'PATCH',
        'QUERY_STRING': 'a=10'
    }
    with Context(environ):
        controller = SimpleJsonPatchController()
        context.form = [
            {'op': 'GET', 'path': '/'},
            {'op': 'PUT', 'path': 'biscuits/1?a=1', 'value': {
                'name':
                'Ginger Nut'
            }},
            {'op': 'GET', 'path': 'biscuits/2', 'value': {
                'name': 'Ginger Nut'
            }},
        ]
        result = ujson.loads(controller())
        assert len(result) == 3




from nanohttp.contexts import context, Context

from restfulpy.principal import ImpersonateAs, DummyIdentity


def test_impersonation():
    with Context({}) as ctx:
        role1 = 'role-1'
        role2 = 'role-2'
        role3 = 'role-3'
        role4 = 'role-4'

        # Simple test
        role_1_principal = DummyIdentity(role1)
        with ImpersonateAs(role_1_principal):
            assert context.identity == role_1_principal
            assert context.identity.is_in_roles(role1)

            # Now we change the role
            role_2_principal = DummyIdentity(role2)
            with ImpersonateAs(role_2_principal):
                assert not context.identity.is_in_roles(role1)
                assert context.identity.is_in_roles(role2)

        # Multiple roles
        role_3_4_principal = DummyIdentity(role3, role4)
        with ImpersonateAs(role_3_4_principal):
            assert context.identity.is_in_roles(role3)
            assert context.identity.is_in_roles(role4)

            assert not context.identity.is_in_roles(role1)
            assert not context.identity.is_in_roles(role2)




import ujson
from sqlalchemy import Unicode, TypeDecorator


class FakeJson(TypeDecorator):  # pragma: no cover
    impl = Unicode

    def process_bind_param(self, value, engine):
        return ujson.dumps(value)

    def process_result_value(self, value, engine):
        if value is None:
            return None

        return ujson.loads(value)



import io
from os import mkdir
from os.path import dirname, abspath, join, exists

from restfulpy.utils import import_python_module_by_filename, \
    construct_class_by_name, copy_stream, md5sum


HERE = abspath(dirname(__file__))
DATA_DIR = join(HERE, 'data')

if not exists(DATA_DIR):
    mkdir(DATA_DIR)


class MyClassToConstructByName:
    def __init__(self, a):
        self.a = a


def test_import_python_module_by_filename():
    filename = join(DATA_DIR, 'a.py')
    with open(filename, mode='w') as f:
        f.write('b = 123\n')

    module_ = import_python_module_by_filename('a', filename)
    assert module_.b == 123

def test_construct_class_by_name():
    obj = construct_class_by_name(
        'restfulpy.tests.test_utils.MyClassToConstructByName',
        1
    )
    assert obj.a ==  1
    assert obj is not None

def test_copy_stream():
    content = b'This is the initial source file'
    source = io.BytesIO(content)
    target = io.BytesIO()
    copy_stream(source, target)
    target.seek(0)
    assert target.read() == content

def test_md5sum():
    content = b'This is the initial source file'
    source = io.BytesIO(content)
    filename = join(DATA_DIR, 'a.txt')
    with open(filename, mode='wb') as f:
        f.write(content)

    assert md5sum(source) == md5sum(filename)




from datetime import datetime

from nanohttp.contexts import Context
from sqlalchemy import Integer, Unicode, DateTime
from sqlalchemy.orm import synonym

from restfulpy.orm import DeclarativeBase, Field, OrderingMixin


class OrderingObject(OrderingMixin, DeclarativeBase):
    __tablename__ = 'orderable_ordering_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))
    _age = Field(Integer)
    created_at = Field(
        DateTime,
        nullable=False,
        json='createdAt',
        readonly=True,
        default=datetime.utcnow,
    )

    def _set_age(self, age):
        self._age = age

    def _get_age(self):  # pragma: no cover
        return self._age

    age = synonym('_age', descriptor=property(_get_age, _set_age))


def test_ordering_mixin(db):
    session =db()

    for i in range(1, 6):
        obj = OrderingObject(
            title=f'object {6-i//2}',
            age=i * 10,
        )
        session.add(obj)

    session.commit()

    query = session.query(OrderingObject)

    # Ascending
    with Context({'QUERY_STRING': 'sort=id'}):
        result = OrderingObject.sort_by_request(query).all()
        assert result[0].id == 1
        assert result[-1].id == 5

    # Descending
    with Context({'QUERY_STRING': 'sort=-id'}):
        result = OrderingObject.sort_by_request(query).all()
        assert result[0].id == 5
        assert result[-1].id == 1

    # Sort by Synonym Property
    with Context({'QUERY_STRING': 'sort=age'}):
        result = OrderingObject.sort_by_request(query).all()
        assert result[0].id == 1
        assert result[-1].id == 5

    # Mutiple sort criteria
    with Context({'QUERY_STRING': 'sort=title,id'}):
        result = OrderingObject.sort_by_request(query).all()
        assert result[0].id == 4
        assert result[1].id == 5

        assert result[2].id == 2
        assert result[3].id == 3

        assert result[4].id == 1

    # Sort by date
    with Context({'QUERY_STRING': 'sort=-createdAt'}):
        result = OrderingObject.sort_by_request(query).all()
        assert result[0].id == 5
        assert result[1].id == 4
        assert result[2].id == 3
        assert result[3].id == 2
        assert result[4].id == 1




from bddrest.authoring import response
from nanohttp import json
from sqlalchemy import Unicode, Integer

from restfulpy.controllers import JsonPatchControllerMixin, ModelRestController
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession, \
    FilteringMixin, PaginationMixin, OrderingMixin, ModifiedMixin
from restfulpy.testing import ApplicableTestCase


class SqlErrorCheckingModel(
    ModifiedMixin,
    FilteringMixin,
    PaginationMixin,
    OrderingMixin,
    DeclarativeBase
):
    __tablename__ = 'sql_error_checking_model'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), unique=True, nullable=False)


class Root(JsonPatchControllerMixin, ModelRestController):
    __model__ = SqlErrorCheckingModel

    @json
    @commit
    def post(self):
        m = SqlErrorCheckingModel()
        m.update_from_request()
        DBSession.add(m)
        return m

    @json
    @SqlErrorCheckingModel.expose
    def get(self, title: str=None):
        query = SqlErrorCheckingModel.query
        if title:
            return query.filter(SqlErrorCheckingModel.title == title)\
                .one_or_none()
        return query


class TestSqlExceptions(ApplicableTestCase):
    __controller_factory__ = Root

    def test_sql_errors(self):
        with self.given(
                'Testing SQL exceptions',
                '/',
                'POST',
                form=dict(title='test')
            ):
            assert response.json['title'] == 'test'

        return
        # unique_violation
        request(
            'ALL',
            'POST',
            '/',
            params=dict(title='test'),
            expected_status='''409 unique_violation ERROR:  duplicate\
            key value violates unique constraint \
            "sql_error_checking_model_title_key"\nDETAIL:\
            Key (title)=(test) already exists.\n''',
        )




from datetime import date, time

import pytest
from nanohttp import HTTPBadRequest, settings
from nanohttp.contexts import Context
from sqlalchemy import Integer, Unicode, ForeignKey, Boolean, Table, Date,\
    Time, Float
from sqlalchemy.orm import synonym

from restfulpy.orm import DeclarativeBase, Field, relationship, composite, \
    ModifiedMixin


class FullName(object):  # pragma: no cover
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def __composite_values__(self):
        return '%s %s' % (self.first_name, self.last_name)

    def __repr__(self):
        return "FullName(%s %s)" % (self.first_name, self.last_name)

    def __eq__(self, other):
        return isinstance(other, FullName) and \
               other.first_name == self.first_name and \
               other.last_name == self.last_name

    def __ne__(self, other):
        return not self.__eq__(other)


class Author(DeclarativeBase):
    __tablename__ = 'author'

    id = Field(Integer, primary_key=True)
    email = Field(
        Unicode(100),
        unique=True,
        index=True,
        json='email',
        pattern=r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)',
        watermark='Email',
        example="user@example.com"
    )
    title = Field(
        Unicode(50),
        index=True,
        min_length=2,
        watermark='First Name'
    )
    first_name = Field(
        Unicode(50),
        index=True,
        json='firstName',
        min_length=2,
        watermark='First Name'
    )
    last_name = Field(
        Unicode(100),
        json='lastName',
        min_length=2,
        watermark='Last Name'
    )
    phone = Field(
        Unicode(10), nullable=True, json='phone', min_length=10,
        pattern=\
            r'\d{3}[-\.\s]??\d{3}[-\.\s]??\d{4}|\(\d{3}\)\s*\d{3}[-\.\s]??'
            r'\d{4}|\d{3}[-\.\s]??\d{4}',
        watermark='Phone'
    )
    name = composite(
        FullName,
        first_name,
        last_name,
        readonly=True,
        json='fullName',
        protected=True
    )
    _password = Field(
        'password',
        Unicode(128),
        index=True,
        json='password',
        protected=True,
        min_length=6
    )
    birth = Field(Date)
    weight = Field(Float(asdecimal=True))
    age = Field(Integer, default=18, minimum=18, maximum=100)

    def _set_password(self, password):
        self._password = 'hashed:%s' % password

    def _get_password(self):  # pragma: no cover
        return self._password

    password = synonym(
        '_password',
        descriptor=property(_get_password, _set_password),
        info=dict(protected=True)
    )


class Memo(DeclarativeBase):
    __tablename__ = 'memo'
    id = Field(Integer, primary_key=True)
    content = Field(Unicode(100), max_length=90)
    post_id = Field(ForeignKey('post.id'), json='postId')


class Comment(DeclarativeBase):
    __tablename__ = 'comment'
    id = Field(Integer, primary_key=True)
    content = Field(Unicode(100), max_length=90)
    special = Field(Boolean, default=True)
    post_id = Field(ForeignKey('post.id'), json='postId')
    post = relationship('Post')


post_tag_table = Table(
    'post_tag', DeclarativeBase.metadata,
    Field('post_id', Integer, ForeignKey('post.id')),
    Field('tag_id', Integer, ForeignKey('tag.id'))
)


class Tag(DeclarativeBase):
    __tablename__ = 'tag'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), watermark='title', label='title')
    posts = relationship(
        'Post',
        secondary=post_tag_table,
        back_populates='tags'
    )


class Post(ModifiedMixin, DeclarativeBase):
    __tablename__ = 'post'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), watermark='title', label='title')
    author_id = Field(ForeignKey('author.id'), json='authorId')
    author = relationship(Author, protected=False)
    memos = relationship(Memo, protected=True, json='privateMemos')
    comments = relationship(Comment, protected=False)
    tags = relationship(
        Tag,
        secondary=post_tag_table,
        back_populates='posts',
        protected=False
    )
    tag_time = Field(Time)


def test_model(db):
    session = db()

    __configuration__ = '''
    timezone:
    '''

    settings.merge(__configuration__)

    with Context({}):
        author1 = Author(
            title='author1',
            email='author1@example.org',
            first_name='author 1 first name',
            last_name='author 1 last name',
            phone=None,
            password='123456',
            birth=date(1, 1, 1),
            weight=1.1
        )

        post1 = Post(
            title='First post',
            author=author1,
            tag_time=time(1, 1, 1)
        )
        session.add(post1)
        session.commit()

        assert post1.id == 1

        post1_dict = post1.to_dict()
        assert {
                'author': {
                    'email': 'author1@example.org',
                    'firstName': 'author 1 first name',
                    'id': 1,
                    'lastName': 'author 1 last name',
                    'phone': None,
                    'title': 'author1',
                    'birth': '0001-01-01',
                    'weight': 1.100,
                    'age': 18
                },
                'authorId': 1,
                'comments': [],
                'id': 1,
                'tags': [],
                'title': 'First post',
                'tagTime': '01:01:01',
            }.items() < post1_dict.items()

        assert 'createdAt' in post1_dict
        assert 'modifiedAt' in post1_dict

        author1_dict = author1.to_dict()
        assert 'fullName' not in author1_dict


def test_metadata(db):
        # Metadata
        author_metadata = Author.json_metadata()
        assert 'id' in author_metadata['fields']
        assert'email' in author_metadata['fields']
        assert author_metadata['fields']['fullName']['protected'] == True
        assert author_metadata['fields']['password']['protected'] == True

        post_metadata = Post.json_metadata()
        assert'author' in post_metadata['fields']

        comment_metadata = Comment.json_metadata()
        assert 'post' in comment_metadata['fields']

        tag_metadata = Tag.json_metadata()
        assert 'posts' in tag_metadata['fields']

        assert Comment.import_value(Comment.__table__.c.special, 'TRUE') ==\
            True




from sqlalchemy import Integer, String

from restfulpy.orm import DeclarativeBase, Field


class Member(DeclarativeBase):
    __tablename__ = 'members'

    id = Field(Integer, primary_key=True, autoincrement=True)
    username = Field(String(50))


def test_db(db):
    session = db()
    assert session.query(Member).count() == 0




import signal
import sys
import threading

from nanohttp import settings

from restfulpy.cli import Launcher, RequireSubCommand


class StartLauncher(Launcher):
    __command__ = 'start'

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            cls.__command__,
            help='Starts the background worker.'
        )

        parser.add_argument(
            '-g',
            '--gap',
            type=int,
            default=None,
            help='Gap between run next task.'
        )

        parser.add_argument(
            '-s',
            '--status',
            default=[],
            action='append',
            help='Task status to process'
        )

        parser.add_argument(
            '-n',
            '--number-of-threads',
            type=int,
            default=None,
            help='Number of working threads'
        )

        parser.add_argument(
            '-f',
            '--filter',
            default=None,
            type=str,
            action='store',
            help='Custom SQL filter for tasks'
        )
        return parser

    def launch(self):

        from restfulpy.taskqueue import worker

        signal.signal(signal.SIGINT, self.kill_signal_handler)
        signal.signal(signal.SIGTERM, self.kill_signal_handler)

        if not self.args.status:
            self.args.status = {'new'}

        if self.args.gap is not None:
            settings.worker.merge({'gap': self.args.gap})

        print(
            f'The following task types would be processed with gap of '
            f'{settings.worker.gap}s:'
        )
        print('Tracking task status(es): %s' % ','.join(self.args.status))

        number_of_threads = \
            self.args.number_of_threads or settings.worker.number_of_threads
        for i in range(number_of_threads):
            t = threading.Thread(
                    target=worker,
                    name='restfulpy-worker-thread-%s' % i,
                    daemon=True,
                    kwargs=dict(
                        statuses=self.args.status,
                        filters=self.args.filter
                    )
                )
            t.start()

        print('Worker started with %d threads' % number_of_threads)
        print('Press Ctrl+C to terminate worker')
        signal.pause()

    @staticmethod
    def kill_signal_handler(signal_number, frame):

        if signal_number == signal.SIGINT:
            print('You pressed Ctrl+C!')
        elif signal_number in (signal.SIGTERM, signal.SIGKILL):
            print('OS Killed Me!')

        sys.stdin.close()
        sys.stderr.close()
        sys.stdout.close()
        sys.exit(1)


class CleanupLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser(
            'cleanup',
            help='Clean database before starting worker processes'
        )

    def launch(self):
        from restfulpy.orm import DBSession
        from restfulpy.taskqueue import RestfulpyTask

        RestfulpyTask.cleanup(DBSession)
        DBSession.commit()


class WorkerLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'worker',
            help="Task queue administration"
        )
        worker_subparsers = parser.add_subparsers(
            title="worker command",
            dest="worker_command"
        )
        StartLauncher.register(worker_subparsers)
        CleanupLauncher.register(worker_subparsers)
        return parser



import pytest
from sqlalchemy import Unicode

from restfulpy.orm import DeclarativeBase, Field, SoftDeleteMixin


class SoftDeleteCheckingModel(SoftDeleteMixin, DeclarativeBase):
    __tablename__ = 'soft_delete_checking_model'

    title = Field(Unicode(50), primary_key=True)


def test_soft_delete_mixin(db):
    session = db()

    instance = SoftDeleteCheckingModel(
        title='test title'
    )
    session.add(instance)
    session.commit()
    instance.assert_is_not_deleted()
    with pytest.raises(ValueError):
        instance.assert_is_deleted()

    instance = session.query(SoftDeleteCheckingModel).one()
    instance.soft_delete()
    session.commit()
    instance.assert_is_deleted()
    with pytest.raises(ValueError):
        instance.assert_is_not_deleted()

    query = session.query(SoftDeleteCheckingModel)
    assert SoftDeleteCheckingModel.filter_deleted(query).count() == 1
    assert SoftDeleteCheckingModel.exclude_deleted(query).count() == 0

    instance.soft_undelete()
    session.commit()
    instance.assert_is_not_deleted()
    with pytest.raises(ValueError):
        instance.assert_is_deleted()

    assert SoftDeleteCheckingModel.filter_deleted(query).count() == 0
    assert SoftDeleteCheckingModel.exclude_deleted(query).count() == 1

    session.delete(instance)
    with pytest.raises(AssertionError):
        session.commit()




from datetime import datetime

from sqlalchemy import Integer, Unicode, not_
from sqlalchemy.sql.expression import desc, asc

from restfulpy.orm import DeclarativeBase, ActivationMixin, PaginationMixin,\
    FilteringMixin, OrderingMixin, Field


class Student(
    DeclarativeBase,
    ActivationMixin,
    PaginationMixin,
    FilteringMixin,
    OrderingMixin
):
    __tablename__ = 'student'
    id = Field(Integer, primary_key=True)
    name = Field(Unicode(100), max_length=90)


def test_activation_mixin(db):
    """
    This unittest is wrote to test the combination of mixins
    """
    session = db()

    activated_student = Student()
    activated_student.name = 'activated-student'
    activated_student.activated_at = datetime.utcnow()
    session.add(activated_student)

    deactivated_student = Student()
    deactivated_student.name = 'deactivated-student'
    deactivated_student.activated_at = None
    session.add(deactivated_student)

    session.commit()

    # Test ordering:
    student_list = session.query(Student).\
        order_by(desc(Student.is_active)).all()
    assert student_list[0].activated_at is not None
    assert student_list[-1].activated_at is None

    student_list = session.query(Student).\
        order_by(asc(Student.is_active)).all()
    assert student_list[-1].activated_at is not None
    assert student_list[0].activated_at is None

    # Test filtering:
    student_list = session.query(Student).filter(Student.is_active).all()
    for student in student_list:
       assert student.activated_at is not None

    student_list = session.query(Student).filter(not_(Student.is_active)).all()
    for student in student_list:
        assert student.activated_at is None




import sys
import argparse
from os import path

import argcomplete

from .cli import Launcher, AutoCompletionLauncher


class RestfulpyMainLauncher(Launcher):

    def __init__(self):
        self.parser = parser = argparse.ArgumentParser(
            description='Restfulpy command line interface.'
        )
        subparsers = parser.add_subparsers(
            title='Restfulpy sub commands',
            prog=path.basename('restfulpy'),
            dest='command'
        )

        AutoCompletionLauncher.register(subparsers)

        from .scaffolding.launchers import ScaffoldLauncher
        ScaffoldLauncher.register(subparsers)

        argcomplete.autocomplete(parser)

    def launch(self, args=None):
        cli_args = self.parser.parse_args(args if args else None)
        if hasattr(cli_args, 'func'):
            cli_args.func(cli_args)
        else:
            self.parser.print_help()
        sys.exit(0)

    @classmethod
    def create_parser(cls, subparsers):
        """
        Do nothing here
        """
        pass


def main():
    RestfulpyMainLauncher()()




from restfulpy.cryptography import AESCipher


def mockup_random(size):
    return b'#' * size


def test_encrypt():
    key = b'1234567890123456'
    message = b'test-message'
    encrypted = b'################>T2\tyZ\\\x0f20[\x97e\x97\xbb5'
    cipher = AESCipher(key, random=mockup_random)
    assert encrypted == cipher.encrypt(message)
    assert message == cipher.decrypt(encrypted)




import os
import sys
from os.path import join, basename, abspath

from .launchers import Launcher, RequireSubCommand


def print_venv_restart_help():
    venv = basename(os.environ['VIRTUAL_ENV'])
    print('\nPlease run this to apply changes auto completion:\n')
    print(f'    deactivate && workon {venv}\n')


class AutoCompletionLauncher(Launcher, RequireSubCommand):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'autocompletion',
            help='Bash autocompletion management'
        )
        sub_subparsers = parser.add_subparsers(dest='autocompletion_command')
        AutoCompletionInstaller.register(sub_subparsers)
        AutoCompletionUninstaller.register(sub_subparsers)
        return parser


class AutoCompletionInstaller(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('install')
        parser.add_argument(
            '-s', '--system-wide',
            action='store_true',
            help=f'Add the PYTHON_ARGCOMPLETE_OK into first 1024 bytes of the  {sys.argv[0]}'
        )
        return parser

    def launch(self):
        if 'VIRTUAL_ENV' in os.environ:
            if self.args.system_wide:
                print(
                    'The -s/--system-wide flag can not be used within virtualenv',
                    file=sys.stderr
                )
                return 1
            self.install_virtualenv()
        elif self.args.system_wide:
            self.install_systemwide()
        else:
            self.install_user()

    def install_virtualenv(self):
        sourcefile = join(os.environ['VIRTUAL_ENV'], 'bin/postactivate')
        result = self.install_file(sourcefile)
        if not result:
            print_venv_restart_help()
        return result

    def install_user(self):
        sourcefile = join(os.environ['HOME'], '.bashrc')
        return self.install_file(sourcefile)

    def install_file(self, filename):
        line = 'eval "$(register-python-argcomplete %s)"\n' % basename(sys.argv[0])

        with open(filename) as f:
            content = f.readlines()

        if line in content:
            print(
                'The autocompletion is already activated.\n'
                f'it means the line:\n\n    {line}\nwas found in file {abspath(filename)}',
                file=sys.stderr
            )
            return 1

        with open(filename, mode='a') as f:
            f.write(line)

        print(f'The line:\n\n    {line}\nwas added into {abspath(filename)}')

    def install_systemwide(self):
        line = '# PYTHON_ARGCOMPLETE_OK'
        filename = sys.argv[0]
        with open(filename) as f:

            content = f.read(1024)
            if line in content:
                print(
                    'The autocompletion is already activated.\n'
                    f'it means the line:\n\n    '
                    f'{line}\nwas found in file {abspath(filename)}',
                    file=sys.stderr
                )
                return 1

            content += f.read()

        lines = content.splitlines()
        lines.insert(1, line)
        with open(filename, mode='w') as f:
            for l in lines:
                f.write(f'{l}\n')

        print(f'The line: {line} was added into: {abspath(filename)}')


class AutoCompletionUninstaller(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('uninstall')
        parser.add_argument(
            '-s', '--system-wide',
            action='store_true',
            help=f'Remove the PYTHON_ARGCOMPLETE_OK from {sys.argv[0]}'
        )
        return parser

    def launch(self):
        if 'VIRTUAL_ENV' in os.environ:
            if self.args.system_wide:
                print(
                    'The -s/--system-wide flag can not be used within ',
                    'virtualenv',
                    file=sys.stderr
                )
                return 1
            self.uninstall_from_virtualenv()
        elif self.args.system_wide:
            self.uninstall_systemwide()
        else:
            self.uninstall_from_user()

    def uninstall_from_virtualenv(self):
        sourcefile = join(os.environ['VIRTUAL_ENV'], 'bin/postactivate')
        result = self.uninstall_from_file(sourcefile)
        if not result:
            print_venv_restart_help()
        return result

    def uninstall_from_user(self):
        sourcefile = join(os.environ['HOME'], '.bashrc')
        return self.uninstall_from_file(sourcefile)

    def uninstall_from_file(self, filename):
        line = \
            f'eval "$(register-python-argcomplete {basename(sys.argv[0])})"\n'
        return self.remove_line_from_file(filename, line)

    def uninstall_systemwide(self):
        line = '# PYTHON_ARGCOMPLETE_OK\n'
        filename = sys.argv[0]
        self.remove_line_from_file(filename, line)

    def remove_line_from_file(self, filename, line):
        with open(filename) as f:
            lines = f.readlines()

        found = False
        with open(filename, mode='w') as f:
            for l in lines:
                if line != l:
                    f.write(l)
                else:
                    found = True

        if found:
            print(
                f'The line:\n\n    '
                f'{line}\nwas removed from {abspath(filename)}'
            )
        else:
            print(
                'The autocompletion is already deactivated.\n'
                f'it means the line:\n\n'
                '    {line}\nwas not found in file {abspath(filename)}',
                file=sys.stderr
            )
            return 1




import argparse
import sys

import argcomplete

from ...cli import Launcher, AutoCompletionLauncher
from .configuration import ConfigurationLauncher
from .database import DatabaseLauncher
from .dev import DevLauncher
from .jwttoken import JWTLauncher
from .migrate import MigrateLauncher
from .serve import ServeLauncher
from .worker import WorkerLauncher
from .erd import EntityRelationshipDiagrams


class MainLauncher(Launcher):

    def __init__(self, application):
        self.application = application
        self.parser = parser = argparse.ArgumentParser(
            prog=application.name,
            description='%s command line interface.' % application.name
        )
        parser.add_argument(
            '-p', '--process-name',
            metavar="PREFIX",
            default=application.name,
            help= \
                'A string indicates the logger prefix for this process, it '
                'helps to configure separate log files per process.'
        )
        parser.add_argument(
            '-c', '--config-file',
            metavar="FILE",
            help='Configuration file, Default: none'
        )
        subparsers = parser.add_subparsers(
            title="sub commands",
            prog=application.name,
            dest="command"
        )

        DatabaseLauncher.register(subparsers)
        ServeLauncher.register(subparsers)
        MigrateLauncher.register(subparsers)
        WorkerLauncher.register(subparsers)
        DevLauncher.register(subparsers)
        AutoCompletionLauncher.register(subparsers)
        EntityRelationshipDiagrams.register(subparsers)

        if application.__configuration_cipher__ is not None:
            ConfigurationLauncher.register(subparsers)

        if application.__authenticator__ is not None:
            JWTLauncher.register(subparsers)

        application.register_cli_launchers(subparsers)
        argcomplete.autocomplete(parser)

    def launch(self, args=None):
        cli_args = self.parser.parse_args(args)
        cli_args.application = self.application
        self.application.process_name = cli_args.process_name
        self.application.configure(files=cli_args.config_file)
        self.application.initialize_orm()
        if hasattr(cli_args, 'func'):
            cli_args.func(cli_args)
        else:
            self.parser.print_help()
        sys.exit(0)

    @classmethod
    def create_parser(cls, subparsers):
        """
        Do nothing here
        """
        pass



from os.path import dirname, abspath

from nanohttp import settings, configure

from restfulpy.messaging.providers import create_messenger, ConsoleMessenger,\
    SmtpProvider


HERE = abspath(dirname(__file__))

def test_messenger_factory():

    __configuration__ = '''
        messaging:
          default_messenger: restfulpy.messaging.ConsoleMessenger
        '''
    configure(force=True)
    settings.merge(__configuration__)

    console_messenger = create_messenger()
    assert isinstance(console_messenger, ConsoleMessenger)

    settings.messaging.default_messenger =\
        'restfulpy.messaging.providers.SmtpProvider'
    smtp_messenger = create_messenger()
    assert isinstance(smtp_messenger, SmtpProvider)




from bddrest.authoring import response
from nanohttp import json, Controller, context

from restfulpy.testing import ApplicableTestCase


class Root(Controller):

    @json
    def index(self):
        return context.form


class TestJSONPayload(ApplicableTestCase):
    __controller_factory__ = Root

    def test_index(self):
        payload = dict(
            a=1,
            b=2
        )

        with self.given(
            'Testing json pyload',
            json=payload,
        ):
            assert response.json ==  payload




from bddrest.authoring import response
from nanohttp import action

from restfulpy.controllers import RootController
from restfulpy.testing import ApplicableTestCase


class Root(RootController):

    @action
    def index(self):
        return 'Index'


class TestApplication(ApplicableTestCase):
    __controller_factory__ = Root

    def test_index(self):
        with self.given('Test application root', '/', 'GET'):
            assert response.status == '200 OK'
            assert response.body == b'Index'

    def test_options(self):
        with self.given('Test OPTIONS verb', '/', 'OPTIONS'):
            assert response.headers['Cache-Control'] == 'no-cache,no-store'




import itsdangerous
from bddrest.authoring import response, when
from nanohttp import json, Controller, context, HTTPBadRequest, settings

from restfulpy.mockup import MockupApplication
from restfulpy.authentication import Authenticator
from restfulpy.authorization import authorize
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.testing import ApplicableTestCase


token_expired = False
refresh_token_expired = False


class MockupMember:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class MockupStatelessAuthenticator(Authenticator):
    def validate_credentials(self, credentials):
        email, password = credentials
        if password == 'test':
            return MockupMember(id=1, email=email, roles=['admin', 'test'])

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(
            id=member_id
        ))

    def create_principal(self, member_id=None, session_id=None):
        return JwtPrincipal(dict(
            id=1,
            email='test@example.com',
            roles=['admin', 'test'],
            sessionId='1')
        )

    def verify_token(self, encoded_token):
        principal = super().verify_token(encoded_token)
        if token_expired:
            raise itsdangerous.SignatureExpired(
                'Simulating',
                payload=principal.payload
            )
        return principal

    def try_refresh_token(self, session_id):
        if refresh_token_expired:
            self.bad()
        return super().try_refresh_token(session_id)


class Root(Controller):
    @json
    def index(self):
        return context.form

    @json(verbs='post')
    def login(self):
        principal = context.application.__authenticator__.login(
            (context.form['email'], context.form['password'])
        )
        if principal:
            return dict(token=principal.dump())
        raise HTTPBadRequest()

    @json
    @authorize
    def me(self):
        return context.identity.payload

    @json()
    @authorize
    def logout(self):
        context.application.__authenticator__.logout()
        return {}

    @json
    @authorize('god')
    def kill(self):  # pragma: no cover
        context.application.__authenticator__.logout()
        return {}


class TestAuthenticator(ApplicableTestCase):
    __application__ = MockupApplication(
        'Authenticator Application',
        Root(),
        authenticator=MockupStatelessAuthenticator()
    )

    __configuration__ = ('''
        jwt:
          max_age: .3
          refresh_token:
            max_age: 3
            secure: true
    ''')

    def test_login(self):
        with self.given(
                'Loggin in to get a token',
                '/login',
                'POST',
                form=dict(email='test@example.com', password='test')
            ):
            assert response.status == '200 OK'
            assert response.headers['X-Identity'] == '1'
            assert 'token' in response.json
            token = response.json['token']

            when(
                'Password is incorrect',
                form=dict(email='test@example.com', password='invalid')
            )
            assert response.status == '400 Bad Request'

        with self.given(
                'Trying to access a protected resource with the token',
                '/',
                'GET',
                form=dict(a='a', b='b'),
                authorization=token
            ):
            assert response.headers['X-Identity'] == '1'

            when('Token is broken', authorization='bad')
            assert response.status == 400

            when('Token is empty', url='/me', authorization='')
            assert response.status == 401

    def test_logout(self):
        self.login(
            dict(
                email='test@example.com',
                password='test'
            ),
            url='/login',
            verb='POST'
        )

        with self.given(
                'Logging out',
                '/logout',
            ):
            assert response.status == '200 OK'
            assert 'X-Identity' not in response.headers

    def test_refresh_token(self):
        global token_expired, refresh_token_expired
        self.logout()
        with self.given(
                'Loggin in to get a token',
                '/login',
                'POST',
                form=dict(email='test@example.com', password='test')
            ):
            refresh_token = response.headers['Set-Cookie'].split('; ')[0]
            assert 'token' in response.json
            assert refresh_token.startswith('refresh-token=')
            token = response.json['token']

            # Login on client
            token_expired = True
            settings.jwt.refresh_token.secure = False


        # Request a protected resource after the token has been expired,
        # with broken cookies
        with self.given(
                    'Refresh token is broken',
                    url='/me',
                    verb='GET',
                    authorization=token,
                    headers={
                        'Cookie': 'refresh-token=broken-data'
                    }
            ):
            assert response.status == 400

            # Request a protected resource after the token has been expired,
            # with empty cookies
            when(
                'Refresh token is empty',
                headers={
                    'Cookie': 'refresh-token'
                }
            )
            assert response.status == 401

            # Request a protected resource after the token has been expired,
            # without the cookies
            when(
                'Without the cookies',
                headers=None
            )
            assert response.status == 401

            # Request a protected resource after the token has been expired,
            # with appropriate cookies.
            when(
                'With appropriate cookies',
                headers={
                    'Cookie': refresh_token
                }
            )
            assert 'X-New-JWT-Token' in response.headers
            assert response.headers['X-New-JWT-Token'] is not None

            # Test with invalid refresh token
            when(
                'With invalid refresh token',
                headers={
                    'Cookie': 'refresh-token=InvalidToken'
                }
            )
            assert response.status == 400

            # Waiting until expire refresh token
            refresh_token_expired = True

            # Request a protected resource after
            # the refresh-token has been expired.
            when(
                'The refresh token has been expired.',
                 headers={
                    'Cookie': 'refresh-token'
                 }
            )
            assert response.status == 401

    def test_authorization(self):
        global token_expired
        self.login(
            dict(
                email='test@example.com',
                password='test'
            ),
            url='/login'
        )
        token_expired = False

        with self.given(
            'Access forbidden',
            url='/kill',
            verb='GET'
        ):
            assert response.status == 403
            assert response.headers['X-Identity'] == '1'




import re
from os.path import join, dirname

from setuptools import setup, find_packages


# reading package version (without reloading it)
with open(join(dirname(__file__), 'restfulpy', '__init__.py')) as v_file:
    package_version = re.compile(r".*__version__ = '(.*?)'", re.S) \
        .match(v_file.read()) \
        .group(1)


dependencies = [
    'nanohttp >= 1.4.4',
    'argcomplete',
    'ujson',
    'appdirs',
    'sqlalchemy',
    'alembic',
    'itsdangerous',
    'mako',
    'psycopg2-binary',
    'redis',
    'pyyaml',
    'ua-parser',
    'user-agents',
    'pycrypto',
    'python-dateutil',

    # Testing
    'requests',
    'pytest',
    'bddrest >= 2'
]


setup(
    name='restfulpy',
    version=package_version,
    description='A toolchain for developing REST APIs',
    author='Vahid Mardani',
    author_email='vahid.mardani@gmail.com',
    install_requires=dependencies,
    packages=find_packages(),
    include_package_data=True,
    test_suite='restfulpy.tests',
    license='MIT',
    entry_points={
        'console_scripts': [
            'restfulpy = restfulpy.launchers:main'
        ]
    },
)



import re
from datetime import datetime, tzinfo, date

from dateutil.parser import parse as dateutil_parse
from dateutil.tz import tzutc, tzstr, tzlocal

from .configuration import settings


POSIX_TIME_PATTERN = re.compile('^\d+(\.\d+)?$')


def localtimezone():
    return tzlocal()


def configuredtimezone():
    timezone = settings.timezone
    if timezone in ('', None):
        return None

    if isinstance(timezone, tzinfo):
        return timezone

    if timezone in (0, 'utc', 'UTC', 'Z', 'z'):
        return tzutc()

    if isinstance(timezone, str):
        return tzstr(timezone)

    raise ValueError(f'Invalid timezone in configuration: {timezone}')


def now():
    timezone = configuredtimezone()
    return datetime.now(timezone)


def parse_datetime(value) -> datetime:
    """Parses a string a a datetime object

    The reason of wrapping this functionality is to preserve compatibility
    and future exceptions handling.

    Another reason is to behave depend to the configuration when parsing date
    and time.
    """
    timezone = configuredtimezone()

    # Parse and return if value is unix timestamp
    if isinstance(value, float) or POSIX_TIME_PATTERN.match(value):
        value = float(value)
        if timezone is None:
            return datetime.fromtimestamp(value, tzutc()) \
                .replace(tzinfo=None)
        else:
            return datetime.fromtimestamp(value, timezone)


    parsed_value = dateutil_parse(value)
    if timezone is not None:
        # The application is configured to use UTC or another time zone:

        # Submit without timezone: Reject and tell the user to specify the
        # timezone.
        if parsed_value.tzinfo is None:
            raise ValueError('You have to specify the timezone')

        # The parsed value is a timezone aware object.
        # If ends with Z: accept and assume as the UTC
        # Then converting it to configured timezone and continue the
        # rest of process
        parsed_value = parsed_value.astimezone(timezone)

    elif parsed_value.tzinfo:
        # The application is configured to use system's local timezone
        # And the sumittd value has tzinfo.
        # So converting it to system's local and removing the tzinfo
        # to achieve a naive object.
        parsed_value = parsed_value\
            .astimezone(localtimezone())\
            .replace(tzinfo=None)

    return parsed_value


def parse_date(value) -> date:
    """Parses a string as a date object

    .. note: It ignores the time part of the value

    The reason of wrapping this functionality is to preserve compatibility
    and future exceptions handling.
    """

    # Parse and return if value is unix timestamp
    if isinstance(value, float) or POSIX_TIME_PATTERN.match(value):
        value = float(value)
        return date.fromtimestamp(value)

    return dateutil_parse(value).date()


def parse_time(value) -> date:
    """Parses a string as a time object

    .. note: It ignores the date part of the value

    The reason of wrapping this functionality is to preserve compatibility
    and future exceptions handling.
    """

    # Parse and return if value is unix timestamp
    if isinstance(value, float) or POSIX_TIME_PATTERN.match(value):
        value = float(value)
        return date.fromtimestamp(value)

    return dateutil_parse(value).time()


def format_datetime(value):
    timezone = configuredtimezone()
    if not isinstance(value, datetime) and isinstance(value, date):
        return value.isoformat()

    if timezone is None and value.tzinfo is not None:
        # The output shoudn't have a timezone specifier.
        # So, converting it to system's local time
        value = value.astimezone(localtimezone()).replace(tzinfo=None)

    elif timezone is not None:
        if value.tzinfo is None:
            raise ValueError('The value must have a timezone specifier')

        elif value.utcoffset() != timezone.utcoffset(value):
            value = value.astimezone(timezone)

    result = value.isoformat()
    if result.endswith('+00:00'):
        result = f'{value[:-6]}Z'

    return result


def format_date(value):
    if isinstance(value, datetime):
        value = datetime.date()

    return value.isoformat()


def format_time(value):
    if isinstance(value, datetime):
        value = datetime.time()

    return value.isoformat()




from os.path import abspath, exists, join, dirname

from appdirs import user_config_dir
from sqlalchemy.exc import SQLAlchemyError

from nanohttp import Application as NanohttpApplication, Controller, \
    HTTPStatus, HTTPInternalServerError, settings
from .cli.main import MainLauncher
from ..authentication import Authenticator
from ..configuration import configure
from ..exceptions import SqlError
from ..logging_ import get_logger
from ..orm import init_model, create_engine, DBSession


class Application(NanohttpApplication):
    """The main entry point

    A web application project should be started by inheritting this class
    and overriding some methods if desirable

    """

    __configuration__ = None
    __logger__ = get_logger()
    __authenticator__ = None
    __configuration_cipher__ = None
    engine = None

    def __init__(self, name: str, root: Controller, root_path='.',
                 version='0.1.0-dev.0', process_name=None, authenticator=None,
                 configuration=None):
        super(Application, self).__init__(root=root)
        self.process_name = process_name or name
        self.version = version
        self.root_path = abspath(root_path)
        self.name = name
        self.cli_main = MainLauncher(self)

        if configuration:
            self.__configuration__ = configuration

        if authenticator:
            self.__authenticator__ = authenticator

        elif self.__authenticator__ is None:
            self.__authenticator__ = Authenticator()

    def _handle_exception(self, ex):
        if isinstance(ex, SQLAlchemyError):
            ex = SqlError(ex)
            self.__logger__.exception(str(ex))
        if not isinstance(ex, HTTPStatus):
            ex = HTTPInternalServerError('Internal server error')
            self.__logger__.exception('Internal server error')
        return super()._handle_exception(ex)

    def configure(self, files=None, context=None, force=False):
        _context = {
            'process_name': self.process_name,
            'root_path': self.root_path,
            'data_dir': join(self.root_path, 'data'),
            'restfulpy_dir': abspath(dirname(__file__))
        }
        if context:
            _context.update(context)

        configure(context=_context, force=force)
        settings.merge(self.__configuration__)

        files = ([files] if isinstance(files, str) else files) or []
        local_config_file = join(user_config_dir(), '%s.yml' % self.name)
        if exists(local_config_file):  # pragma: no cover
            files.insert(0, local_config_file)

        for filename in files:
            with open(filename, 'rb') as f:
                header = f.read(4)
                if header == b'#enc':
                    content = self.__configuration_cipher__.decrypt(f.read())
                else:
                    content = header + f.read()
                settings.merge(content.decode())

    def register_cli_launchers(self, subparsers):
        """
        This is a template method
        """
        pass

    def initialize_orm(self, engine=None):
        if engine is None:
            engine = create_engine()

        self.engine = engine
        init_model(engine)

    # Hooks
    def begin_request(self):
        self.__authenticator__.authenticate_request()

    def end_response(self):
        DBSession.remove()

    def insert_basedata(self):  # pragma: no cover
        raise NotImplementedError()

    def insert_mockup(self):  # pragma: no cover
        raise NotImplementedError()

    def shutdown(self):
        DBSession.close_all()
        self.engine.dispose()

